# restate/restate/backends/cached.py
from __future__ import annotations

import atexit
import asyncio
import time
from pathlib import PurePosixPath as Path
from typing_extensions import (
    Any,
    Generic,
    Literal,
    NotRequired,
    TypeAlias,
    TypeVar,
    TypedDict,
    Union,
    Unpack,
)

from restate.shared.sentinel import Sentinel

from .base import Backend, AsyncBackend
from .memory import InMemoryBackend


_B = TypeVar("_B", bound=Union[Backend, AsyncBackend])
_T = TypeVar("_T")

OperationType: TypeAlias = Literal["read", "write", "delete"]


class Operation:
    def __init__(self, op_type: OperationType, path: Path, value: Any | None = None):
        self.type = op_type
        self.path = path
        self.value = value
        self.timestamp = time.time()


class FlushArgs(TypedDict):
    flush_interval: NotRequired[float]
    flush_on_read: NotRequired[bool]
    flush_on_write: NotRequired[bool]
    flush_on_delete: NotRequired[bool]


class CachingBackendBase(Generic[_B]):
    """Base class for caching backends with common functionality"""

    def __init__(self, backend: _B, **kwargs: Unpack[FlushArgs]):
        flush_interval = kwargs.get("flush_interval", 5.0)
        flush_on_read = kwargs.get("flush_on_read", False)
        flush_on_write = kwargs.get("flush_on_write", True)
        flush_on_delete = kwargs.get("flush_on_delete", True)

        self.backend = backend
        self.cache = InMemoryBackend()
        self.flush_interval = flush_interval

        self.flush_triggers: dict[OperationType, bool] = {
            "read": flush_on_read,
            "write": flush_on_write,
            "delete": flush_on_delete,
        }

        self.operations: list[Operation] = []
        self.last_flush = time.time()

    def schedule_operation(
        self,
        op_type: OperationType,
        path: Path,
        value: Any | None = None,
    ) -> bool:
        self.operations.append(Operation(op_type, path, value))

        return self.check_flush(op_type)

    def check_flush(self, op_type: OperationType) -> bool:
        if not self.flush_triggers[op_type]:
            return False

        return time.time() - self.last_flush >= self.flush_interval


class CachingSyncBackend(Backend, CachingBackendBase[Backend]):
    def __init__(self, backend: Backend, **kwargs: Unpack[FlushArgs]):
        super().__init__(backend, **kwargs)
        atexit.register(self.flush)

    def read(
        self,
        path: Path,
        default: _T = None,
    ) -> Any | _T:
        fake_default = Sentinel("fake_default")

        value = self.cache.read(path, default=fake_default)

        if value is fake_default:
            value = self.backend.read(path, default)
            self.cache.write(path, value)

        if self.check_flush("read"):
            self.flush()

        return value

    def write(
        self,
        path: Path,
        value: Any | None,
    ) -> None:
        self.cache.write(path, value)

        if self.schedule_operation("write", path, value):
            self.flush()

    def delete(self, path: Path) -> None:
        self.cache.delete(path)

        if self.schedule_operation("delete", path):
            self.flush()

    def flush(self) -> None:
        for operation in self.operations:
            if operation.type == "write":
                self.backend.write(operation.path, operation.value)
            elif operation.type == "delete":
                self.backend.delete(operation.path)

        self.operations.clear()
        self.last_flush = time.time()


class CachingAsyncBackend(AsyncBackend, CachingBackendBase[AsyncBackend]):
    def __init__(self, backend: AsyncBackend, **kwargs: Unpack[FlushArgs]):
        super().__init__(backend, **kwargs)
        atexit.register(self._sync_flush)

    async def read(
        self,
        path: Path,
        default: _T = None,
    ) -> Any | _T:
        fake_default = Sentinel("fake_default")

        value = self.cache.read(path, default=fake_default)

        if value is fake_default:
            value = await self.backend.read(path, default)
            self.cache.write(path, value)

        if self.check_flush("read"):
            await self.flush()

        return value

    async def write(
        self,
        path: Path,
        value: Any | None,
    ) -> None:
        self.cache.write(path, value)

        if self.schedule_operation("write", path, value):
            await self.flush()

    async def delete(self, path: Path) -> None:
        self.cache.delete(path)

        if self.schedule_operation("delete", path):
            await self.flush()

    async def flush(self) -> None:
        operations = self.operations
        self.operations = []
        self.last_flush = time.time()

        for operation in operations:
            if operation.type == "write":
                await self.backend.write(operation.path, operation.value)
            elif operation.type == "delete":
                await self.backend.delete(operation.path)

    def _sync_flush(self):
        """Synchronous flush for cleanup on exit"""
        loop = asyncio.new_event_loop()
        try:
            loop.run_until_complete(self.flush())
        finally:
            loop.close()



from __future__ import annotations

from pathlib import PurePath
from typing_extensions import Any, TypeAlias, TypeVar
from restate.shared.constants import ROOT_PATH
from restate.shared.sentinel import Sentinel
from .base import Backend


NestedStore: TypeAlias = dict

NESTED_KEY = Sentinel("nested")

_T = TypeVar("_T")


class InMemoryBackend(Backend):
    """
    In-memory storage.
    Fast, responsive, but loses all state on restart.
    """

    def __init__(self):
        self.stores: dict[PurePath, Store] = {ROOT_PATH: Store(self, ROOT_PATH)}

    def get_or_create_store(self, path: PurePath) -> Store:
        if path not in self.stores:
            self.stores[path] = Store(self, path)

            parent_store = self.get_or_create_store(path.parent)
            parent_store.children.add(path)

        return self.stores[path]

    def get_store(self, path: PurePath) -> Store | None:
        return self.stores.get(path)

    def read(
        self,
        path: PurePath,
        default: _T = None,
    ) -> Any | _T:
        store = self.get_store(path)

        if not store:
            return default

        return store.data

    def write(self, path: PurePath, value: Any | None) -> None:
        store = self.get_or_create_store(path)
        store.set_value(value)

    def delete(self, path: PurePath) -> None:
        store = self.get_store(path)

        if not store:
            return

        for child_path in store.children:
            self.delete(child_path)

        if path != ROOT_PATH:  # do not delete root
            self.stores.pop(path)

            parent_store = self.get_store(path.parent)

            if parent_store:
                parent_store.children.remove(path)


class Store:
    def __init__(self, backend: InMemoryBackend, path: PurePath) -> None:
        self.path = path
        self.children: set[PurePath] = set()
        self.backend = backend
        self._has_value: bool = False
        self._value: Any = None

    def set_value(self, value: Any) -> None:
        self._has_value = True
        self._value = value

    @property
    def data(self):
        if not self.children:
            if self._has_value:
                return self._value

            return {}

        resulting_data = {}

        if self._has_value:
            resulting_data["/"] = self._value

        for child_path in self.children:
            data = self.backend.read(child_path)

            if child_path.is_relative_to(self.path):
                child_path = child_path.relative_to(self.path)

            resulting_data[str(child_path)] = data

        return resulting_data

    def __repr__(self) -> str:
        fields = {
            "path": self.path,
            "value": self._value if self._has_value else "-",
            "children": len(self.children),
        }

        return f"Store({', '.join(f'{k}: {v}' for k,v in fields.items())})"



from typing_extensions import Any, Self, Sequence, Callable, TypeVar
import operator
from pathlib import PurePosixPath as Path

from restate.atoms.sync_atom import SyncAtom

from .internal_types import EqualityFunction, PathLike, StateCallback, StateEvent
from .base import BaseController, DeriveData
from .callback_store import CallbackID

from restate.shared.constants import ROOT_PATH
from restate.shared.sentinel import Sentinel
from restate.backends.base import Backend
from restate.backends.memory import InMemoryBackend


_T = TypeVar("_T")


fake_default = Sentinel("state_default")


class ControllerSync(BaseController):
    def __init__(self, backend: Backend | None = None):
        if backend is None:
            backend = InMemoryBackend()

        self.backend: Backend = backend

        super().__init__()

    def get_state(
        self,
        path: PathLike,
        default: _T = None,
        write_default: bool = False,
    ) -> Any | _T:
        """
        Read the state at the `path`

        Pass `default` as a value to return if state doesn't exist at `path`.
        Pass `write_default=True` to write `default` value if state doesn't exist at `path`.
        """

        value = self.backend.read(self.resolve_path(path), fake_default)

        if isinstance(value, Sentinel):
            if write_default:
                self.write_state(path, default)

            return default

        return value

    def write_state(
        self,
        path: PathLike,
        value: Any | None,
    ):
        self.backend.write(self.resolve_path(path), value)

    def notify(
        self,
        path: PathLike,
        event: StateEvent[Self, _T],
    ):
        path = self.resolve_path(path)

        for callback_id in self.callbacks.get_callbacks(path):
            # if event is bubbling, notify every caller
            # otherwise notify only special-cased callers
            if event.bubbling or isinstance(callback_id, Sentinel):
                self.callbacks.call_sync(callback_id, event)

        if path != ROOT_PATH:
            self.notify(path.parent, event.get_bubbled())

    def set_state(
        self,
        path: PathLike,
        value: Any | None,
        eq_func: EqualityFunction | None = None,
        default: Any | None = None,
        payload: Any = None,
        skip_notify: bool = False,
    ) -> bool:
        """
        Sets the path state to the given value and notifies the subscribers.
        If value is the same as the previous one, doesn't notify or rewrite.
        The comparison behaviour is done by the given eq_func (`==` comparison by default)

        `default` is passed to get_state to retrieve the prev_value.

        If you want to notify all subscribers unconditionally, check .ping method (it skips writing and comparison)

        Returns boolean to show if a rewrite has happened
        """

        path = self.resolve_path(path)
        prev_value = self.get_state(path, default)

        if not eq_func:
            eq_func = operator.eq

        are_equal = bool(eq_func(value, prev_value))

        if are_equal:
            return False

        event = self.build_event(path, prev_value, value, payload)

        self.write_state(path, value)

        if not skip_notify:
            self.notify(path, event)

        return True

    def del_state(
        self,
        path: PathLike,
        payload: Any = None,
        skip_notify: bool = False,
    ) -> bool:
        path = self.resolve_path(path)
        prev_value = self.get_state(path, fake_default)

        if prev_value == fake_default:
            return False

        event = self.build_event(path, prev_value, None, payload)

        self.backend.delete(path)

        if not skip_notify:
            self.notify(path, event)

        return True

    def ping(
        self,
        path: PathLike,
        default: _T | None = None,
        payload: Any = None,
    ):
        """
        Notifies path subscribers unconditionally.
        default is passed to .get_state
        """

        value: _T | None = self.get_state(path, default)

        event = self.build_event(path, value, value, payload)

        self.notify(path, event)

    def derive_many(
        self,
        dest: PathLike,
        sources: Sequence[PathLike],
        transform: Callable[[DeriveData], Any | None],
        payload: Any = None,
    ) -> CallbackID:
        def callback(event: StateEvent):
            update_data: dict[Path, Any | None] = {}

            for p in sources:
                resolved = self.resolve_path(p)

                if resolved == event.emitting_path:
                    update_data[resolved] = event.new_value
                else:
                    update_data[resolved] = self.get_state(p)

            self.set_state(
                dest, transform(DeriveData(self, update_data)), payload=payload
            )

        callback_id = Sentinel(f"derive:{dest}")
        self.callbacks.remove_callback(callback_id)
        self.register_callback(callback, force_id=callback_id)

        start_event = self.build_event(ROOT_PATH, None, None, payload)

        for p in sources:
            self.subscribe_by_id(
                p,
                callback_id,
            )

        callback(start_event)

        return callback_id

    def derive(
        self,
        dest: PathLike,
        source: PathLike,
        transform: Callable[[Any | None], Any | None],
        payload: Any = None,
    ) -> CallbackID:
        return self.derive_many(
            dest,
            [source],
            transform=lambda d: transform(d.get(source)),
            payload=payload,
        )

    def track(self, callback: StateCallback[Self, _T], payload: Any = None):
        callback_id = self.register_callback(callback)

        def wrapped_callback(event: StateEvent[Self, _T]):
            tracker = tracker_controller.create_tracker()
            event.tracker = tracker
            result = self.callbacks.call_sync(callback_id, event)
            event.tracker = None
            tracker_controller.submit_tracker(tracker)
            return result

        wrapped_id = self.register_callback(wrapped_callback)

        tracker_controller = StateTrackerController(self, wrapped_callback)

        self.callbacks.call_sync(
            wrapped_id,
            self.build_event(
                ROOT_PATH,
                prev_value=None,
                new_value=None,
                payload=payload,
            ),
        )

    def atom(
        self,
        path: PathLike,
        default: _T | None = None,
    ) -> SyncAtom[_T]:
        return SyncAtom(
            controller=self,
            path=path,
            default=default,
        )


from .tracker import StateTrackerController  # noqa: E402



from __future__ import annotations

from collections.abc import Iterable
from pathlib import PurePosixPath
from typing_extensions import Literal, TypeAlias
from restate.shared.caller import FunctionAdapter
from restate.shared.sentinel import Sentinel


CallbackID: TypeAlias = str | Sentinel


def is_glob_path(path: PurePosixPath) -> bool:
    for part in path.parts:
        if part == "*":
            return True

    return False


class CallbackStore:
    def __init__(self, asyncify_strategy: Literal["thread", "none"] = "none"):
        self.subscriptions: dict[PurePosixPath, set[CallbackID]] = {}
        self.callbacks: dict[CallbackID, StateCallback] = {}
        self.reverse_subscriptions: dict[CallbackID, set[PurePosixPath]] = {}
        self.asyncify_strategy = asyncify_strategy
        self.glob_paths: set[PurePosixPath] = set()

    def get_id(self, callback_or_id: CallbackID | StateCallback) -> CallbackID:
        if isinstance(callback_or_id, (str, Sentinel)):
            return callback_or_id

        return f"#{id(callback_or_id):02x}"

    def add_callback(
        self,
        callback: StateCallback,
        force_id: CallbackID | None = None,
        replace: bool = False,
    ) -> CallbackID:
        callback_id = self.get_id(force_id or callback)

        while True:  # single-iteration loop to do breaks
            if callback_id in self.callbacks:
                if callback == self.callbacks[callback_id]:
                    return callback_id

                if not replace:
                    raise ValueError(
                        f"Cannot add callback '{callback_id}': another with the same id already exists. If you want to replace state, pass replace=True to store.add_callback"
                    )

            break

        self.callbacks[callback_id] = callback

        return callback_id

    def get_matching_paths(self, path: PurePosixPath) -> Iterable[PurePosixPath]:
        if path in self.subscriptions:
            yield path

        parts = path.parts

        target_len = len(parts)

        for glob_path in self.glob_paths:
            if len(glob_path.parts) != target_len:
                continue

            for i, part in enumerate(glob_path.parts):
                if part == "*":
                    continue

                if parts[i] != part:
                    break

            else:
                yield glob_path

    def get_callbacks(self, path: PurePosixPath) -> Iterable[CallbackID]:
        for matching_path in self.get_matching_paths(path):
            for callback_id in list(self.subscriptions.get(matching_path, [])):
                yield callback_id

    def subscribe(
        self,
        path: PurePosixPath,
        callback_id: CallbackID,
        ignore_missing: bool = False,
    ):
        if is_glob_path(path):
            self.glob_paths.add(path)

        if callback_id not in self.callbacks and not ignore_missing:
            raise ValueError(
                "CallbackStore.subscribe received an unknown callback_id. If you intend to add the callback later, pass ignore_missing=True"
            )

        if path not in self.subscriptions:
            self.subscriptions[path] = set()

        if callback_id not in self.reverse_subscriptions:
            self.reverse_subscriptions[callback_id] = set()

        self.subscriptions[path].add(callback_id)
        self.reverse_subscriptions[callback_id].add(path)

    def unsubscribe(self, path: PurePosixPath, callback_id: CallbackID):
        if path not in self.subscriptions:
            return

        if callback_id not in self.subscriptions:
            return

        sub_set = self.subscriptions[path]

        if callback_id in sub_set:
            sub_set.remove(callback_id)

        if not sub_set:
            self.subscriptions.pop(path)

        rev_sub_set = self.reverse_subscriptions[callback_id]

        if path in rev_sub_set:
            rev_sub_set.remove(path)

        if not rev_sub_set:
            if is_glob_path(path):
                self.glob_paths.remove(path)

            self.reverse_subscriptions.pop(callback_id)

    def remove_callback(
        self,
        callback_id: CallbackID,
    ) -> StateCallback | None:
        rev_subs = self.reverse_subscriptions.pop(callback_id, ())

        for path in rev_subs:
            self.unsubscribe(path, callback_id)

        return self.callbacks.pop(callback_id, None)

    async def call_async(
        self,
        callback_id: CallbackID,
        event: StateEvent,
    ):
        callback = self.callbacks.get(callback_id)

        if callback is None:
            return

        adapted = FunctionAdapter(callback)  # type: ignore because Pyright is mad, 'function is not assignable to AnyFunction'
        await adapted.call_async(event)

    def call_sync(
        self,
        callback_id: CallbackID,
        event: StateEvent,
    ):
        callback = self.callbacks.get(callback_id)

        if callback is None:
            return

        # TODO: raise an issue in Pyright
        adapted = FunctionAdapter(callback)  # type: ignore because Pyright is mad, 'function is not assignable to AnyFunction'
        adapted.call_sync(event)


from .internal_types import StateCallback, StateEvent  # noqa: E402



from __future__ import annotations
from typing import TypeVar

from typing_extensions import (
    Any,
    Callable,
    Self,
)

from pathlib import PurePosixPath as Path


_T = TypeVar("_T")


class BaseController:
    def __init__(self):
        self.callbacks: CallbackStore = CallbackStore()

    def resolve_path(self, path: PathLike) -> Path:
        while not isinstance(path, Path):
            if isinstance(path, str):
                if not path:
                    path = "/"

                path = Path(path)
            else:
                path = path.path

        if path.parts[:1] != ("/"):
            path = Path("/") / path

        return path

    def register_callback(
        self,
        callback: StateCallback[Self],
        force_id: CallbackID | None = None,
        replace: bool = False,
    ):
        """
        Registers a callback for further use. A part of register-subscribe process.
        You may also use .subscribe method directly if you have both path and callback on hands.
        Returns the callback_id to be used in subscriptions.

        Pass a string force_id to set a specific callback_id.
        If `replace=False`, passing a different callback with the same force_id will result in an error.
        Pass `replace=True` if you intend to overwrite the callback.
        """

        return self.callbacks.add_callback(callback, force_id, replace)

    def subscribe_by_id(
        self,
        path: PathLike,
        callback_id: CallbackID,
        ignore_missing: bool = False,
    ) -> CallbackID:
        """
        Pass a callback_id instead of a real callback.
        Allows to decouple callback creation and subscription.

        If called before callback creation, will raise errors unless `ignore_missing=True`
        """
        self.callbacks.subscribe(
            path=self.resolve_path(path),
            callback_id=callback_id,
            ignore_missing=ignore_missing,
        )

        return callback_id

    def subscribe(
        self,
        path: PathLike,
        callback: StateCallback[Self],
        force_id: CallbackID | None = None,
        replace: bool = False,
    ) -> CallbackID:
        """
        Subscribes a callback to every path update (including all bubbling updates of child paths).
        To subscribe with a callback_id, use .subscribe_by_id.
        To get a callback_id without any subscriptions, use .register_callback

        Pass a string force_id to set a specific callback_id.
        If `replace=False`, passing a different callback with the same force_id will result in an error.
        Pass `replace=True` if you intend to overwrite the callback.
        """

        callback_id = self.register_callback(callback, force_id, replace)
        self.subscribe_by_id(path, callback_id)
        return callback_id

    def unsubscribe_by_id(
        self,
        path: PathLike,
        callback_id: CallbackID,
    ):
        """
        Remove a subscription for the given path and callback_id
        """
        self.callbacks.unsubscribe(self.resolve_path(path), callback_id)

    def unsubscribe(
        self,
        path: PathLike,
        callback: StateCallback[Self],
    ):
        """
        Remove a subscription for the given path and callback
        """
        callback_id = self.callbacks.get_id(callback)
        self.unsubscribe_by_id(path, callback_id)

    def build_event(
        self,
        path: PathLike,
        prev_value: _T | None,
        new_value: _T | None,
        payload: Any,
    ) -> StateEvent[Self, _T]:
        path = self.resolve_path(path)

        return StateEvent(
            controller=self,
            emitting_path=path,
            current_path=path,
            prev_value=prev_value,
            new_value=new_value,
            payload=payload,
        )

    def get_state(
        self,
        path: PathLike,
        default: Any | None = None,
        write_default: bool = False,
    ) -> Any: ...

    def set_state(
        self,
        path: PathLike,
        value: Any | None,
        eq_func: Callable[[Any | None, Any | None], bool] | None = None,
        default: Any | None = None,
        payload: Any = None,
        skip_notify: bool = False,
    ) -> bool: ...

    def del_state(
        self,
        path: PathLike,
        payload: Any = None,
        skip_notify: bool = False,
    ) -> bool: ...


class DeriveData:
    def __init__(self, controller: BaseController, data: dict[Path, Any]):
        self.controller = controller
        self.data = {self.controller.resolve_path(p): v for p, v in data.items()}

    def get(self, path: PathLike) -> Any | None:
        path = self.controller.resolve_path(path)

        return self.data.get(path)

    def __getattr__(self, attr: str):
        return self.get(attr)


from .internal_types import PathLike, StateCallback, StateEvent  # noqa: E402
from .callback_store import CallbackID, CallbackStore  # noqa: E402



from pathlib import PurePosixPath as Path
from typing_extensions import Any, TypeVar
from restate.backends.base import AsyncBackend, Backend


_T = TypeVar("_T")


class AsyncifyBackend(AsyncBackend):
    """
    This backend "converts" any sync backend into async one.
    This will not magically make your blocking calls non-blocking, this is just interface compat.
    Used internally by async controller.
    """

    def __init__(self, sync_backend: Backend):
        self.sync_backend = sync_backend

    async def read(
        self,
        path: Path,
        default: _T = None,
    ) -> Any | _T:
        return self.sync_backend.read(path, default)

    async def write(self, path: Path, value: Any | None) -> None:
        return self.sync_backend.write(path, value)

    async def delete(self, path: Path) -> None:
        return self.sync_backend.delete(path)



from __future__ import annotations

from dataclasses import dataclass
from pathlib import PurePosixPath
from time import perf_counter
from typing_extensions import Any, Generic, TypeVar


AnyController = TypeVar(
    "AnyController",
    bound="BaseController",
    covariant=True,
)


@dataclass
class StateTracker(Generic[AnyController]):
    controller: AnyController
    run_start: float = -1
    run_end: float = -1
    _used_paths: set[PurePosixPath] | None = None

    @property
    def paths(self):
        if self._used_paths is None:
            self._used_paths = set()

        return self._used_paths

    def start(self):
        self.run_start = perf_counter()

    def stop(self):
        self.run_end = perf_counter()

    def track(self, path: PurePosixPath):
        self.paths.add(path)

    def get_state(
        self,
        path: PurePosixPath | str,
        default: Any | None = None,
        write_default: bool = False,
    ):
        self.track(self.controller.resolve_path(path))
        return self.controller.get_state(path, default, write_default)


class StateTrackerController(Generic[AnyController]):
    def __init__(
        self, state_controller: AnyController, callback: StateCallback[AnyController]
    ):
        self.last_run: StateTracker | None = None
        self.callback = callback
        self.state_controller = state_controller

    def create_tracker(self):
        return StateTracker(self.state_controller)

    def reconcile(
        self,
        old_run: StateTracker[AnyController] | None,
        new_run: StateTracker[AnyController],
    ):
        if old_run is None:
            old_paths = set()
        else:
            old_paths = old_run.paths

        new_paths = new_run.paths

        for path in old_paths | new_paths:
            if path not in new_paths:
                self.state_controller.unsubscribe(path, self.callback)

            if path not in old_paths:
                self.state_controller.subscribe(path, self.callback)

    def submit_tracker(
        self,
        run: StateTracker[AnyController],
    ):
        if self.last_run:
            if self.last_run.run_end > run.run_end:
                return

        self.reconcile(self.last_run, run)

        self.last_run = run


from .internal_types import StateCallback  # noqa: E402
from .base import BaseController  # noqa: E402



from __future__ import annotations
from typing import TYPE_CHECKING, Any, Callable, TypeVar


from restate.controllers.callback_store import CallbackID
from restate.controllers.internal_types import (
    EqualityFunction,
    PathLike,
    StateCallback,
)

if TYPE_CHECKING:
    from restate.controllers import ControllerSync
    from restate.controllers.base import DeriveData

from .core import BaseAtom

_T = TypeVar("_T")
_R = TypeVar("_R")


class SyncAtom(BaseAtom[_T]):
    def __init__(
        self,
        controller: ControllerSync,
        path: PathLike,
        default: _T = None,
    ) -> None:
        self.default = default
        self.controller = controller
        self.path = path

    def set(
        self,
        value: _T,
        eq_func: EqualityFunction | None = None,
        payload: Any = None,
        skip_notify: bool = False,
    ):
        self.controller.set_state(
            self.path,
            value,
            eq_func=eq_func,
            default=self.default,
            payload=payload,
            skip_notify=skip_notify,
        )

    def get(self, write_default: bool = False) -> _T:
        return self.controller.get_state(
            self.path,
            default=self.default,
            write_default=write_default,
        )

    def subscribe_by_id(
        self,
        callback_id: CallbackID,
        ignore_missing: bool = False,
    ) -> CallbackID:
        return self.controller.subscribe_by_id(
            self.path,
            callback_id,
            ignore_missing=ignore_missing,
        )

    def subscribe(
        self,
        callback: StateCallback[ControllerSync, Any],
        force_id: CallbackID | None = None,
        replace: bool = False,
    ) -> CallbackID:
        return self.controller.subscribe(
            self.path,
            callback,
            force_id=force_id,
            replace=replace,
        )

    def derive(
        self,
        dest: PathLike | BaseAtom[_R],
        transform: Callable[[_T], _R],
    ):
        return self.controller.derive(
            dest=dest,
            source=self.path,
            transform=transform,  # type: ignore (atom default)
        )

    def ping(
        self,
        payload: Any = None,
    ):
        self.controller.ping(
            self.path,
            self.default,
            payload,
        )

    def derive_from(
        self,
        *sources: PathLike,
        transform: Callable[[DeriveData], _T],
        payload: Any = None,
    ):
        self.controller.derive_many(
            self.path,
            sources,
            transform,
            payload=payload,
        )



from __future__ import annotations

from pathlib import PurePosixPath as Path
from typing_extensions import Any, TypeVar

_T = TypeVar("_T")


class Backend:
    def read(
        self,
        path: Path,
        default: _T = None,
    ) -> Any | _T: ...

    def write(
        self,
        path: Path,
        value: Any | None,
    ) -> None: ...

    def delete(self, path: Path) -> None: ...


class AsyncBackend:
    async def read(
        self,
        path: Path,
        default: Any | _T = None,
    ) -> Any | _T: ...

    async def write(
        self,
        path: Path,
        value: Any | None,
    ) -> None: ...

    async def delete(self, path: Path) -> None: ...



import asyncio
from dataclasses import dataclass
from pathlib import Path, PurePosixPath
from typing_extensions import Any, Callable, Generic, Iterable, Literal, TypeVar
import json
import os
import aiofiles
import aiofiles.os as aio_os
import shutil

from restate.shared.constants import ROOT_PATH
from .base import Backend, AsyncBackend

_R = TypeVar("_R", str, bytes)
_C = TypeVar("_C", default=Any)
_T = TypeVar("_T")


@dataclass
class Serializer(Generic[_R, _C]):
    extension: str
    raw_type: type[_R]
    deserialize: Callable[[_R], _C]
    serialize: Callable[[_C], _R]

    @property
    def extension_fallback(self):
        fallback = ["bin", "txt"][self.raw_type is str]

        if not self.extension:
            return fallback

        return self.extension

    @property
    def suffix(self):
        return f".{self.extension_fallback}"

    @property
    def file_mode(self) -> Literal["b", "t"]:
        if self.raw_type is str:
            return "t"
        return "b"

    @property
    def write_mode(self) -> Literal["wb", "wt"]:
        return f"w{self.file_mode}"  # type: ignore

    @property
    def read_mode(self) -> Literal["rb", "rt"]:
        return f"r{self.file_mode}"  # type: ignore


json_serializer = Serializer(
    extension="json",
    raw_type=str,
    serialize=json.dumps,
    deserialize=json.loads,
)


class FileSystemBackendBase(Generic[_R]):
    def __init__(self, base_path: str | Path, serializer: Serializer[_R]):
        self.base_path = Path(base_path).absolute()
        self.serializer = serializer

    def get_fs_path(self, path: PurePosixPath) -> Path:
        relative = path.relative_to(ROOT_PATH)
        return self.base_path / relative

    def get_children_dir(self, path: PurePosixPath):
        return self.get_fs_path(path)

    def get_doc_path(self, path: PurePosixPath):
        return self.get_fs_path(path).with_suffix(self.serializer.suffix)

    def get_file_mode(self, write: bool = False):
        prefix = "rw"[write]

        if self.serializer.raw_type is str:
            return prefix

        return f"{prefix}b"


class FileSystemSyncBackend(FileSystemBackendBase, Backend):
    """
    Synchronous filesystem storage backend.

    Stores data in a directory structure with serialized files. Each path maps to both
    a possible document file and a directory for children. For example, for JSON serializer:

    ```
    /                        -> /
        users/               -> /users
            1.json           -> /users/1
            settings.json    -> /users/settings
            groups/          -> /users/groups
                admin.json   -> /users/groups/admin
    ```

    Features:
    - Configurable serialization format (JSON by default)
    - Automatic directory creation
    - Nested path support

    Args:
        base_path: Root directory for storage
        serializer: Serializer instance defining data format (defaults to JSON)

    Example:
        >>> backend = FileSystemSyncBackend("./storage")
        >>> backend.write("/users/1", {"name": "John"})
        >>> backend.read("/users/1")
        {'name': 'John'}
        >>> backend.read("/users")
        {'1': {'name': 'John'}}
    """

    def __init__(
        self,
        base_path: str | Path,
        serializer: Serializer[_R] = json_serializer,
    ):
        super().__init__(base_path, serializer)

    def ensure_parent_dirs(self, path: Path) -> None:
        path.parent.mkdir(parents=True, exist_ok=True)

    def iter_children(self, path: PurePosixPath) -> Iterable[PurePosixPath]:
        children_dir = self.get_children_dir(path)

        if not children_dir.exists():
            return

        visited: set[PurePosixPath] = set()

        for filepath in children_dir.iterdir():
            if (filepath.suffix == self.serializer.suffix) or (
                filepath.suffix == "" and filepath.is_dir()
            ):
                child_path = path / filepath.name
            else:
                continue

            if child_path not in visited:
                visited.add(child_path)
                yield child_path.with_suffix("")

    def read(
        self,
        path: PurePosixPath,
        default: _T = None,
    ) -> Any | _T:
        doc_path = self.get_doc_path(path)

        resulting_data = {}

        if doc_path.exists():
            with open(doc_path, self.serializer.read_mode) as f:
                raw_data = f.read()

            resulting_data["/"] = self.serializer.deserialize(raw_data)

        for child_path in self.iter_children(path):
            resulting_data[child_path.name] = self.read(child_path)

        if not resulting_data:
            return default

        if len(resulting_data) == 1 and "/" in resulting_data:
            return resulting_data["/"]

        return resulting_data

    def write(
        self,
        path: PurePosixPath,
        value: Any | None,
    ) -> None:
        doc_path = self.get_doc_path(path)
        self.ensure_parent_dirs(doc_path)

        serialized = self.serializer.serialize(value)

        with open(doc_path, self.serializer.write_mode) as f:
            f.write(serialized)

    def delete(self, path: PurePosixPath) -> None:
        doc_path = self.get_doc_path(path)
        dir_path = self.get_children_dir(path)

        try:
            os.remove(doc_path)
        except Exception:
            pass

        try:
            shutil.rmtree(dir_path)
        except Exception:
            pass

        parent_folder = doc_path.parent

        while parent_folder != self.base_path:
            try:
                parent_folder.rmdir()
            except OSError:  # directory not empty
                return
            else:
                parent_folder = parent_folder.parent


class FileSystemAsyncBackend(FileSystemBackendBase, AsyncBackend):
    """
    Asynchronous filesystem storage backend.

    Identical to FileSystemSyncBackend in structure and features, but provides
    async interface using aiofiles for non-blocking I/O operations.

    Storage structure example for JSON serializer:

    ```
    /                        -> /
        users/               -> /users
            1.json           -> /users/1
            settings.json    -> /users/settings
            groups/          -> /users/groups
                admin.json   -> /users/groups/admin
    ```

    Features:
    - Configurable serialization format (JSON by default)
    - Automatic directory creation
    - Nested path support

    Args:
        base_path: Root directory for storage
        serializer: Serializer instance defining data format (defaults to JSON)

    Example:
        >>> backend = FileSystemAsyncBackend("./storage")
        >>> await backend.write("/users/1", {"name": "John"})
        >>> await backend.read("/users/1")
        {'name': 'John'}
        >>> await backend.read("/users")
        {'1': {'name': 'John'}}
    """

    def __init__(
        self,
        base_path: str | Path,
        serializer: Serializer[_R] = json_serializer,
    ):
        super().__init__(base_path, serializer)

    async def ensure_parent_dirs(self, path: Path) -> None:
        await aio_os.makedirs(path.parent, exist_ok=True)

    async def iter_children(self, path: PurePosixPath) -> Iterable[PurePosixPath]:
        children_dir = self.get_children_dir(path)

        if not await aio_os.path.exists(children_dir):
            return []

        children = await aio_os.listdir(children_dir)

        visited: set[PurePosixPath] = set()

        for filepath_str in children:
            filepath = Path(filepath_str)

            if (filepath.suffix == self.serializer.suffix) or (
                filepath.suffix == "" and filepath.is_dir()
            ):
                child_path = path / filepath.name
            else:
                continue

            if child_path not in visited:
                visited.add(child_path)

        return visited

    async def read(
        self,
        path: PurePosixPath,
        default: _T = None,
    ) -> Any | _T:
        doc_path = self.get_doc_path(path)

        resulting_data = {}

        if await aio_os.path.exists(doc_path):
            async with aiofiles.open(doc_path, self.serializer.read_mode) as f:
                raw_data = await f.read()

            resulting_data["/"] = self.serializer.deserialize(raw_data)

        for child_path in await self.iter_children(path):
            resulting_data[child_path.name] = self.read(child_path)

        if not resulting_data:
            return default

        if len(resulting_data) == 1 and "/" in resulting_data:
            return resulting_data["/"]

        return resulting_data

    async def write(
        self,
        path: PurePosixPath,
        value: Any | None,
    ) -> None:
        doc_path = self.get_doc_path(path)
        await self.ensure_parent_dirs(doc_path)

        async with aiofiles.open(doc_path, self.serializer.write_mode) as f:
            await f.write(self.serializer.serialize(value))

    async def delete(self, path: PurePosixPath) -> None:
        doc_path = self.get_doc_path(path)
        dir_path = self.get_children_dir(path)

        try:
            await aio_os.remove(doc_path)
        except Exception:
            pass

        try:
            await asyncio.to_thread(
                shutil.rmtree,
                dir_path,
            )
        except Exception:
            pass

        parent_folder = doc_path.parent

        while parent_folder != self.base_path:
            try:
                await aio_os.rmdir(parent_folder)
            except OSError:  # directory not empty
                return
            else:
                parent_folder = parent_folder.parent



from .base import Backend as Backend, AsyncBackend as AsyncBackend
from .memory import InMemoryBackend as InMemoryBackend
from .asyncify import AsyncifyBackend as AsyncifyBackend
from .hybrid import (
    HybridAsyncBackend as HybridAsyncBackend,
    HybridSyncBackend as HybridSyncBackend,
)
from .cached import (
    CachingAsyncBackend as CachingAsyncBackend,
    CachingSyncBackend as CachingSyncBackend,
)
from .fs import (
    FileSystemAsyncBackend as FileSystemAsyncBackend,
    FileSystemSyncBackend as FileSystemSyncBackend,
)



from __future__ import annotations

from pathlib import PurePosixPath as Path
from typing import Callable
from typing_extensions import Any, Awaitable, Self, Sequence, TypeVar
import operator

from .base import BaseController, DeriveData
from .callback_store import CallbackID

from restate.shared.constants import ROOT_PATH
from restate.backends.asyncify import AsyncifyBackend
from restate.backends.base import AsyncBackend, Backend
from restate.backends.memory import InMemoryBackend
from restate.shared.sentinel import Sentinel

from restate.atoms import AsyncAtom

_T = TypeVar("_T")


fake_default = Sentinel("state_default")


class ControllerAsync(BaseController):
    def __init__(
        self,
        backend: AsyncBackend | Backend | None = None,
    ):
        if backend is None:
            backend = InMemoryBackend()

        if not isinstance(backend, AsyncBackend):
            backend = AsyncifyBackend(backend)

        self.backend: AsyncBackend = backend

        super().__init__()

    async def get_state(
        self,
        path: PathLike,
        default: _T = None,
        write_default: bool = False,
    ) -> Any | _T:
        """
        Read the state at the `path`

        Pass `default` as a value to return if state doesn't exist at `path`.
        Pass `write_default=True` to write `default` value if state doesn't exist at `path`.
        """
        value = await self.backend.read(self.resolve_path(path), fake_default)

        if isinstance(value, Sentinel):
            if write_default:
                await self.write_state(path, default)

            return default

        return value

    async def write_state(
        self,
        path: PathLike,
        value: Any | None,
    ):
        await self.backend.write(self.resolve_path(path), value)

    async def notify(
        self,
        path: PathLike,
        event: StateEvent[Self, _T],
    ):
        path = self.resolve_path(path)

        for callback_id in self.callbacks.get_callbacks(path):
            # if event is bubbling, notify every caller
            # otherwise notify only special-cased callers
            if event.bubbling or isinstance(callback_id, Sentinel):
                await self.callbacks.call_async(callback_id, event)

        if path != ROOT_PATH:
            await self.notify(path.parent, event.get_bubbled())

    async def set_state(  # pyright: ignore[reportIncompatibleMethodOverride]
        self,
        path: PathLike,
        value: Any | None,
        eq_func: EqualityFunction | None = None,
        default: Any | None = None,
        payload: Any = None,
        skip_notify: bool = False,
    ) -> bool:
        """
        Sets the path state to the given value and notifies the subscribers.
        If value is the same as the previous one, doesn't notify or rewrite.
        The comparison behaviour is done by the given eq_func (`==` comparison by default)

        `default` is passed to get_state to retrieve the prev_value.

        If you want to notify all subscribers unconditionally, check .ping method (it skips writing and comparison)

        Returns boolean to show if a rewrite has happened.
        """

        path = self.resolve_path(path)
        prev_value = await self.get_state(path, default)

        if not eq_func:
            eq_func = operator.eq

        are_equal = bool(eq_func(value, prev_value))

        if are_equal:
            return False

        event = self.build_event(path, prev_value, value, payload)

        await self.write_state(path, value)

        if not skip_notify:
            await self.notify(path, event)

        return True

    async def del_state(  # pyright: ignore[reportIncompatibleMethodOverride]
        self,
        path: PathLike,
        payload: Any = None,
        skip_notify: bool = False,
    ) -> bool:
        path = self.resolve_path(path)
        prev_value = await self.get_state(path, fake_default)

        if prev_value == fake_default:
            return False

        event = self.build_event(path, prev_value, None, payload)

        await self.backend.delete(path)

        if not skip_notify:
            await self.notify(path, event)

        return True

    async def ping(
        self,
        path: PathLike,
        default: _T | None = None,
        payload: Any = None,
    ):
        """
        Notifies path subscribers unconditionally.
        default is passed to .get_state
        """

        value: _T | None = await self.get_state(path, default)

        await self.notify(path, self.build_event(path, value, value, payload))

    async def derive_many(
        self,
        dest: PathLike,
        sources: Sequence[PathLike],
        transform: Callable[
            [DeriveData],
            Awaitable[Any | None],
        ],
        payload: Any = None,
    ) -> CallbackID:
        async def callback(event: StateEvent[Self]):
            update_data: dict[Path, Any | None] = {}

            for p in sources:
                resolved = self.resolve_path(p)

                if p == event.emitting_path:
                    update_data[resolved] = event.new_value
                else:
                    update_data[resolved] = await self.get_state(p)

            await self.set_state(
                dest,
                await transform(DeriveData(self, update_data)),
                payload=payload,
            )

        callback_id = Sentinel(f"derive:{dest}")
        self.callbacks.remove_callback(callback_id)
        self.register_callback(callback, force_id=callback_id)

        start_event = self.build_event(ROOT_PATH, None, None, payload)

        for p in sources:
            self.subscribe_by_id(
                p,
                callback_id,
            )

        await callback(start_event)

        return callback_id

    async def derive(
        self,
        dest: PathLike,
        source: PathLike,
        transform: Callable[[Any | None], Awaitable[Any | None]],
        payload: Any = None,
    ) -> CallbackID:
        def full_transform(d: DeriveData) -> Awaitable[Any | None]:
            return transform(d.get(source))

        return await self.derive_many(
            dest,
            [source],
            transform=full_transform,
            payload=payload,
        )

    async def track(self, callback: StateCallback[Self, _T], payload: Any = None):
        callback_id = self.register_callback(callback)

        async def wrapped_callback(event: StateEvent[Self, _T]):
            tracker = tracker_controller.create_tracker()
            event.tracker = tracker
            await self.callbacks.call_async(callback_id, event)
            event.tracker = None
            tracker_controller.submit_tracker(tracker)

        wrapped_id = self.register_callback(wrapped_callback)

        tracker_controller = StateTrackerController(self, wrapped_callback)

        await self.callbacks.call_async(
            wrapped_id,
            self.build_event(
                ROOT_PATH,
                prev_value=None,
                new_value=None,
                payload=payload,
            ),
        )

    def atom(self, path: PathLike, default: _T | None = None) -> AsyncAtom[_T]:
        return AsyncAtom(
            controller=self,
            path=path,
            default=default,
        )


from .internal_types import EqualityFunction, PathLike, StateEvent, StateCallback  # noqa: E402
from .tracker import StateTrackerController  # noqa: E402



from __future__ import annotations
from pathlib import PurePosixPath as Path

from dataclasses import dataclass
from typing import cast

from typing_extensions import (
    Any,
    Callable,
    Coroutine,
    Generic,
    Protocol,
    TypeAlias,
    TypeVar,
    overload,
)


type EqualityFunction = Callable[[Any, Any], bool]


AnyController = TypeVar(
    "AnyController",
    bound="BaseController | AsyncControllerProtocol",
    covariant=True,
)

_T = TypeVar("_T")

type PathLike = PathLikeClass | Path | str


class PathLikeClass(Protocol):
    path: PathLike


class AsyncControllerProtocol(Protocol):
    async def get_state(
        self,
        path: Path | str,
        default: Any | None = None,
        write_default: bool = False,
    ) -> Any: ...

    async def set_state(
        self,
        path: Path | str,
        value: Any | None,
        eq_func: Callable[[Any | None, Any | None], bool] | None = None,
        default: Any | None = None,
        payload: Any = None,
    ) -> bool: ...

    async def del_state(
        self,
        path: Path | str,
        payload: Any = None,
        skip_notify: bool = False,
    ) -> bool: ...


_T = TypeVar("_T", default=Any)


@dataclass
class StateEvent(Generic[AnyController, _T]):
    controller: AnyController
    emitting_path: Path
    current_path: Path
    prev_value: _T | None
    new_value: _T | None
    bubbling: bool = True
    tracker: StateTracker[BaseController] | None = None
    payload: Any = None

    def get_bubbled(self) -> StateEvent[AnyController]:
        return StateEvent(
            controller=self.controller,
            emitting_path=self.emitting_path,
            current_path=self.emitting_path.parent,
            prev_value=self.prev_value,
            new_value=self.new_value,
            bubbling=self.bubbling,
            tracker=self.tracker,
        )

    @overload
    def get_state(
        self: StateEvent[AsyncControllerProtocol, _T],
        path: Path | str,
        default: _T = None,
        write_default: bool = False,
    ) -> Coroutine[Any, Any, Any | _T]: ...

    @overload
    def get_state(
        self: StateEvent[AnyController, _T],
        path: Path | str,
        default: Any | _T = None,
        write_default: bool = False,
    ) -> Any | _T: ...

    def get_state(
        self,
        path: Path | str,
        default: Any | None = None,
        write_default: bool = False,
    ):
        if self.tracker:
            return self.tracker.get_state(path, default, write_default)

        return self.controller.get_state(path, default, write_default)

    @overload
    def set_state(
        self: StateEvent[AsyncControllerProtocol],
        path: Path | str,
        value: Any | None,
        eq_func: Callable[[Any | None, Any | None], bool] | None = None,
        default: Any | None = None,
        payload: Any = None,
        skip_notify: bool = False,
    ) -> Coroutine[Any, Any, bool]: ...

    @overload
    def set_state(
        self: StateEvent[AnyController],
        path: Path | str,
        value: Any | None,
        eq_func: EqualityFunction | None = None,
        default: Any | None = None,
        payload: Any = None,
        skip_notify: bool = False,
    ) -> bool: ...

    def set_state(
        self,
        path: Path | str,
        value: Any | None,
        eq_func: EqualityFunction | None = None,
        default: Any | None = None,
        payload: Any = None,
        skip_notify: bool = False,
    ) -> Coroutine[Any, Any, bool] | bool:
        return cast(
            Any,
            self.controller.set_state(
                path=path,
                value=value,
                eq_func=eq_func,
                default=default,
                payload=payload,
            ),
        )

    @overload
    def del_state(
        self: StateEvent[AsyncControllerProtocol, _T],
        path: Path | str,
        payload: Any,
        skip_notify: bool = False,
    ) -> Coroutine[Any, Any, bool]: ...

    @overload
    def del_state(
        self: StateEvent[AnyController, _T],
        path: Path | str,
        payload: Any,
        skip_notify: bool = False,
    ) -> bool: ...

    def del_state(
        self,
        path: Path | str,
        payload: Any,
        skip_notify: bool = False,
    ):
        return self.controller.del_state(
            path,
            payload,
            skip_notify,
        )


SyncCallback = Callable[[StateEvent[AnyController, _T]], Any]
AsyncCallback = Callable[[StateEvent[AnyController, _T]], Coroutine[Any, Any, Any]]


StateCallback: TypeAlias = (
    SyncCallback[AnyController, _T] | AsyncCallback[AnyController, _T]
)


from .tracker import StateTracker  # noqa: E402
from .base import BaseController  # noqa: E402



class SingletonMeta(type):
    _instances = {}

    def __call__(cls, identifier: str):
        if identifier not in cls._instances:
            cls._instances[identifier] = super().__call__(identifier)
        return cls._instances[identifier]


class Sentinel(metaclass=SingletonMeta):
    def __init__(self, identifier: str):
        self.identifier = identifier

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Sentinel):
            return False

        return self.identifier == other.identifier

    def __repr__(self) -> str:
        return f"Sentinel({self.identifier})"

    def __hash__(self):
        return hash(self.identifier)


known: dict[str, Sentinel] = {}



from .backends import (
    InMemoryBackend as InMemoryBackend,
    AsyncifyBackend as AsyncifyBackend,
    HybridAsyncBackend as HybridAsyncBackend,
    HybridSyncBackend as HybridSyncBackend,
    CachingAsyncBackend as CachingAsyncBackend,
    CachingSyncBackend as CachingSyncBackend,
    FileSystemAsyncBackend as FileSystemAsyncBackend,
    FileSystemSyncBackend as FileSystemSyncBackend,
)

from .backends.fs import Serializer as Serializer

from .controllers import (
    ControllerSync as ControllerSync,
    ControllerAsync as ControllerAsync,
    StateEvent as StateEvent,
)

from .controllers.base import DeriveData as DeriveData

from .shared.constants import ROOT_PATH as ROOT_PATH



from __future__ import annotations

import asyncio
from asyncio.protocols import Protocol
from typing_extensions import Coroutine, Generic, ParamSpec, TypeIs, TypeVar


_P = ParamSpec("_P")
_R = TypeVar("_R")


class SyncFunction(Protocol, Generic[_P, _R]):
    def __call__(self, *args: _P.args, **kwds: _P.kwargs) -> _R: ...


class AsyncFunction(Protocol, Generic[_P, _R]):
    def __call__(
        self, *args: _P.args, **kwds: _P.kwargs
    ) -> Coroutine[None, None, _R]: ...


AnyFunction = SyncFunction[_P, _R] | AsyncFunction[_P, _R]


def is_awaitable(
    cb: _R | Coroutine[None, None, _R],
) -> TypeIs[Coroutine[None, None, _R]]:
    return asyncio.iscoroutine(cb)


class FunctionAdapter(Generic[_P, _R]):
    def __init__(
        self,
        func: AnyFunction[_P, _R],
    ):
        self.func = func

    async def call_async(
        self,
        *args: _P.args,
        **kwargs: _P.kwargs,
    ) -> _R:
        maybe_coro = self.func(*args, **kwargs)

        if is_awaitable(maybe_coro):
            return await maybe_coro

        return maybe_coro

    def call_sync(
        self,
        *args: _P.args,
        **kwargs: _P.kwargs,
    ) -> _R:
        maybe_coro = self.func(*args, **kwargs)

        if is_awaitable(maybe_coro):
            loop = asyncio.get_event_loop()
            return loop.run_until_complete(maybe_coro)

        return maybe_coro



from pathlib import PurePosixPath as Path
from typing_extensions import Any, Generic, Union, TypeVar

from restate.shared.constants import ROOT_PATH
from .base import Backend, AsyncBackend


_B = TypeVar("_B", bound=Union[Backend, AsyncBackend])
_T = TypeVar("_T")


class HybridBackendBase(Generic[_B]):
    def __init__(self, root_backend: _B):
        self.mounts: dict[Path, _B] = {}
        self.mount(ROOT_PATH, root_backend)

    def resolve_path(self, path: Path | str) -> Path:
        return ROOT_PATH / str(path).lstrip("/")

    def mount(self, path: Path | str, backend: _B):
        path = self.resolve_path(path)

        self.mounts[path] = backend

    def unmount(self, path: Path | str) -> _B | None:
        path = self.resolve_path(path)

        return self.mounts.pop(path, None)

    def resolve_backend(self, path: Path) -> tuple[Path, _B]:
        viable_paths = [
            mount_path for mount_path in self.mounts if path.is_relative_to(mount_path)
        ]

        viable_paths.sort(key=lambda path: len(path.parts), reverse=True)

        best_mount = viable_paths[0]

        return (
            ROOT_PATH / path.relative_to(best_mount),
            self.mounts[best_mount],
        )

    def get_subpaths(self, root_path: Path) -> list[Path]:
        subpaths: list[Path] = []

        for mount in self.mounts:
            if mount == root_path:
                continue

            if mount.is_relative_to(root_path):
                subpaths.append(mount)

        # this bit is important
        # we should work with shortest paths first to not overwrite children
        return sorted(subpaths, key=lambda p: len(p.parts))


class HybridSyncBackend(Backend, HybridBackendBase[Backend]):
    def read(
        self,
        path: Path,
        default: _T = None,
    ) -> Any | _T:
        local_path, backend = self.resolve_backend(path)
        base = backend.read(local_path, default)

        subpaths = self.get_subpaths(path)

        if isinstance(base, dict):
            for subpath in subpaths:
                store = base

                data = self.read(subpath)

                if data is None:
                    continue

                path_parts = list(
                    filter(
                        None,
                        (part.strip("/") for part in subpath.relative_to(path).parts),
                    )
                )

                if not path_parts:
                    # mounted path is the same as local?, we just rewrite the base
                    # shouldn't happen normally but

                    if isinstance(data, dict):
                        base = data

                    continue

                for part in path_parts[:-1]:
                    if not part.strip("/"):
                        continue

                    store[part] = {}
                    store = store[part]

                store[path_parts[-1]] = data

        return base

    def write(
        self,
        path: Path,
        value: Any | None,
    ) -> None:
        local_path, backend = self.resolve_backend(path)
        return backend.write(local_path, value)

    def delete(self, path: Path) -> None:
        local_path, backend = self.resolve_backend(path)
        return backend.delete(local_path)


class HybridAsyncBackend(AsyncBackend, HybridBackendBase[AsyncBackend]):
    async def read(
        self,
        path: Path,
        default: _T = None,
    ) -> Any | _T:
        local_path, backend = self.resolve_backend(path)
        base = await backend.read(local_path, default)

        if isinstance(base, dict):
            subpaths = self.get_subpaths(path)

            for subpath in subpaths:
                store = base

                data = await self.read(subpath)

                if data is None:
                    continue

                path_parts = list(
                    filter(
                        None,
                        (part.strip("/") for part in subpath.relative_to(path).parts),
                    )
                )

                if not path_parts:
                    # mounted path is the same as local?, we just rewrite the base
                    # shouldn't happen normally but

                    if isinstance(data, dict):
                        base = data

                    continue

                for part in path_parts[:-1]:
                    if not part.strip("/"):
                        continue

                    store[part] = {}
                    store = store[part]

                store[path_parts[-1]] = data

        return base

    async def write(
        self,
        path: Path,
        value: Any | None,
    ) -> None:
        local_path, backend = self.resolve_backend(path)
        return await backend.write(local_path, value)

    async def delete(self, path: Path) -> None:
        local_path, backend = self.resolve_backend(path)
        return await backend.delete(local_path)
